import { Comment, DeclarationReflection, LiteralType, ParameterReflection, ReferenceType, ReflectionFlags, ReflectionKind, ReflectionType, SignatureReflection, SomeType, TupleType, TypeOperatorType } from 'typedoc';
import { AllowedHttpMethod, ParentReflection } from '../model';
import { NAME_EXTERNAL_DRIVER, NAME_TYPES_MODULE } from './builtin-external-driver';
import { NAME_BUILTIN_COMMAND_MODULE, NAME_METHOD_MAP } from './builtin-method-map';
import { NAME_NEW_METHOD_MAP, NAME_EXECUTE_METHOD_MAP, NAME_PARAMS } from './external';
declare type WithName<S extends string, R> = R & {
    name: S;
};
declare type WithKind<K extends ReflectionKind, R> = R & {
    kind: K;
};
/**
 * Utility to narrow a declaration reflection to a specific `SomeType`
 */
declare type WithSomeType<T extends SomeType, R> = R & {
    type: T;
};
/**
 * Utility to narrow by name and kind
 */
declare type WithNameAndKind<S extends string, K extends ReflectionKind, R> = R & {
    name: S;
    kind: K;
};
/**
 * Utility; a TupleType with literal elements
 */
export declare type TupleTypeWithLiteralElements = TupleType & {
    elements: LiteralType[];
};
/**
 * Type corresponding to a reflection of a {@linkcode @appium/types#MethodMap}
 */
export declare type MethodMapDeclarationReflection = WithName<typeof NAME_METHOD_MAP | typeof NAME_NEW_METHOD_MAP, DeclarationReflectionWithReflectedType>;
/**
 * Type corresponding to a reflection of {@linkcode @appium/base-driver}
 */
export declare type BaseDriverDeclarationReflection = WithNameAndKind<typeof NAME_BUILTIN_COMMAND_MODULE, ReflectionKind.Module, DeclarationReflection>;
/**
 * Type for the parameters of a command definition or execute method definition.
 *
 * Node that merging `TypeOperatorType` won't work because it will no longer satisfy `SomeType`, because `SomeType` is a finite collection.
 */
export declare type MethodDefParamNamesDeclarationReflection = WithSomeType<TypeOperatorTypeWithTupleTypeWithLiteralElements, DeclarationReflection>;
export declare type TypeOperatorTypeWithTupleTypeWithLiteralElements = TypeOperatorType & {
    operator: 'readonly';
    target: TupleTypeWithLiteralElements;
};
/**
 * Narrows a declaration reflection to one having a reflection type and a property kind. Generic
 */
export declare type PropDeclarationReflection = WithKind<ReflectionKind.Property, DeclarationReflectionWithReflectedType>;
/**
 * A type corresponding to the HTTP method of a route, which is a property off of the object with the route name in a `MethodMap`
 */
export declare type HTTPMethodDeclarationReflection = WithName<AllowedHttpMethod, PropDeclarationReflection>;
/**
 * A declaration reflection having a reflection type. Generic
 */
export declare type DeclarationReflectionWithReflectedType = WithSomeType<ReflectionType, DeclarationReflection>;
/**
 * Type corresponding to the value of the `command` property within a `MethodDef`, which must be a type literal.
 */
export declare type CommandPropDeclarationReflection = WithSomeType<LiteralType, DeclarationReflection>;
/**
 * A generic type guard
 */
export declare type Guard<T> = (value: any) => value is T;
/**
 * Type corresponding to an execute method map
 */
export declare type ExecMethodDeclarationReflection = WithName<typeof NAME_EXECUTE_METHOD_MAP, DeclarationReflectionWithReflectedType> & WithStaticFlag;
/**
 * Whatever has this flag will be a static member
 */
export declare type WithStaticFlag = {
    flags: ReflectionFlags & {
        isStatic: true;
    };
};
/**
 * Whatever has this flag will _not_ be a static member
 */
export declare type WithoutStaticFlag = {
    flags: ReflectionFlags & {
        isStatic: false;
    };
};
/**
 * Type corresponding to the `params` prop of a `MethodDef`
 */
export declare type MethodDefParamsPropDeclarationReflection = WithNameAndKind<typeof NAME_PARAMS, ReflectionKind.Property, DeclarationReflectionWithReflectedType>;
/**
 * Type corresponding to the `payloadParams` prop of an `ExecMethodDef`
 */
export declare type ExecMethodDefParamsPropDeclarationReflection = WithName<typeof NAME_PARAMS, DeclarationReflectionWithReflectedType>;
/**
 * Type corresponding to `@appium/types` module
 */
export declare type AppiumTypesReflection = WithNameAndKind<typeof NAME_TYPES_MODULE, ReflectionKind.Module, ParentReflection>;
/**
 * Type corresponding to a TS `interface`
 */
export declare type InterfaceDeclarationReflection = WithKind<ReflectionKind.Interface, DeclarationReflection>;
/**
 * Type corresponding to the `ExternalDriver` `interface` of `@appium/types`
 */
export declare type ExternalDriverDeclarationReflection = WithName<typeof NAME_EXTERNAL_DRIVER, InterfaceDeclarationReflection>;
/**
 * A call signature for a function that returns some sort of `Promise`.
 */
export declare type AsyncCallSignatureReflection = CallSignatureReflection & {
    type: ReferenceType;
    name: 'Promise';
};
/**
 * An async method or reference to an async method.  In a driver, a command's method must be of this type.
 */
export declare type CommandMethodDeclarationReflection<T extends ReferenceType | ReflectionType = ReferenceType> = WithSomeType<T, DeclarationReflection> & WithKind<T extends ReferenceType ? ReflectionKind.Method : ReflectionKind.Property, DeclarationReflection> & WithoutStaticFlag & (T extends ReferenceType ? {
    signatures: NonEmptyArray<AsyncCallSignatureReflection>;
} : T extends ReflectionType ? {
    type: {
        declaration: {
            signatures: NonEmptyArray<AsyncCallSignatureReflection>;
        };
    };
} : never);
/**
 * A little "struct" containing a method and potentially a derived comment (which may not have been derived from
 * the method itself).
 */
export interface KnownMethodData {
    comment?: Comment;
    method: CommandMethodDeclarationReflection;
}
/**
 * A lookup of command names to {@linkcode KnownMethodData} objects.
 */
export declare type KnownMethods = Map<string, KnownMethodData>;
/**
 * A {@linkcode DeclarationReflection} which is a `class`.
 */
export declare type ClassDeclarationReflection = WithKind<ReflectionKind.Class, DeclarationReflection>;
/**
 * A constructor
 */
export declare type ConstructorDeclarationReflection = WithNameAndKind<'constructor', ReflectionKind.Constructor, DeclarationReflection>;
/**
 * A {@linkcode ReferenceType} referencing the constructor of `BasePlugin`
 */
export declare type BasePluginConstructorReferenceType = ReferenceType & {
    name: 'BasePlugin.constructor';
};
/**
 * A {@linkcode DeclarationReflection} for the constructor of a class extending `BasePlugin`
 */
export declare type BasePluginConstructorDeclarationReflection = WithSomeType<ReferenceType, DeclarationReflection> & ConstructorDeclarationReflection & ({
    inheritedFrom: BasePluginConstructorReferenceType;
} | {
    overwrites: BasePluginConstructorReferenceType;
});
/**
 * One of {@linkcode ExecMethodDefParamsPropDeclarationReflection} or
 * {@linkcode MethodDefParamsPropDeclarationReflection}, which are "parameters" properties of method
 * definition objects (as in a `MethodMap`) or execute method definitions (in an `ExecMethodMap`)
 */
export declare type ParamsPropDeclarationReflection = ExecMethodDefParamsPropDeclarationReflection | MethodDefParamsPropDeclarationReflection;
/**
 * A {@linkcode SignatureReflection} which is a call signature; a function signature.
 *
 * (Other types of signatures include things like "constructor signatures")
 */
export declare type CallSignatureReflection = WithKind<ReflectionKind.CallSignature, SignatureReflection>;
/**
 * An array with a nonzero number of items.
 */
export declare type NonEmptyArray<T> = [T, ...T[]];
/**
 * A {@linkcode CallSignatureReflection} with a nonzero number of parameters.
 *
 * This is used to rename parameters on commands to prefer the ones as defined in the method map.
 */
export declare type CallSignatureReflectionWithArity = CallSignatureReflection & {
    parameters: NonEmptyArray<ParameterReflection>;
};
export {};
//# sourceMappingURL=types.d.ts.map